structure Type :> Type =
struct

  (* Use "raise Error (message,position)" for error messages *)
  exception Error of string*(int*int)

  datatype JanusType = Integer | Array of int

  type pos = int*int

  (* lookup function for symbol table as list of (name,value) pairs *)
  fun lookup x []
        = NONE
    | lookup x ((y,v)::table)
        = if x=y then SOME v else lookup x table

  (* ADDED: specific function for looking up variables in Def lists - needed in checkOutDefs *)
  fun lookupDef x []
        = NONE
    | lookupDef x (Janus.IntVarDef (y,pos)::table)
        = if x=y then SOME Integer else lookupDef x table
    | lookupDef x (Janus.ArrayVarDef (y,size,pos)::table)
        = if x=y then SOME (Array size) else lookupDef x table


  (* check expression *)
  fun checkExp exp vtable avoid =
  (* avoid is the LHS variable that can not be used on the RHS *)
    case exp of
      Janus.Num (n,pos) => ()
    | Janus.LVal lv =>
       (case lv of
           Janus.IntVar (x,p) =>
	     (case lookup x vtable of
               SOME Integer =>
                 if x=avoid
                 then raise Error ("LHS variable used on RHS",p)
	         	 else ()
	      | _ => raise Error ("Array variable used as Integer",p))
        |  Janus.ArrayIndex (x,e,p) =>
	     (case lookup x vtable of
			SOME (Array _) =>
              if x=avoid
              then raise Error ("LHS array used in index or on RHS",p)
	          else checkExp e vtable avoid
	      | _ => raise Error ("Integer variable used as Array",p)))
    | Janus.Plus (e1,e2,pos) =>
       (checkExp e1 vtable avoid; checkExp e2 vtable avoid)
    | Janus.Minus (e1,e2,pos) =>
       (checkExp e1 vtable avoid; checkExp e2 vtable avoid)
    | Janus.Half (e1,pos) => checkExp e1 vtable avoid

  (* check condition *)
  fun checkCond cond vtable =
    case cond of
      Janus.Equal (e1,e2,pos) =>
       (checkExp e1 vtable ""; checkExp e2 vtable "")
    | Janus.Less (e1,e2,pos) =>
       (checkExp e1 vtable ""; checkExp e2 vtable "")
    | Janus.Not (c1,pos) => checkCond c1 vtable
    | Janus.And (c1,c2,pos) =>
       (checkCond c1 vtable; checkCond c2 vtable)
    | Janus.Or (c1,c2,pos) =>
       (checkCond c1 vtable; checkCond c2 vtable)
	| Janus.Empty pos => ()  (* ADDED for stack - no need for type checking here *)

  fun checkStat s vtable pnames =
    case s of
      Janus.Sequence (s1,s2,pos) =>
        (checkStat s1 vtable pnames; checkStat s2 vtable pnames)
    | Janus.AddUpdate (lv,e,pos) =>
        (case lv of
           Janus.IntVar (x,p) =>
             (case lookup x vtable of
               SOME Integer =>
                 checkExp e vtable x
	      	  | _ => raise Error ("array variable used as integer",p))
        |  Janus.ArrayIndex (x,e1,p) =>
             (case lookup x vtable of
				SOME (Array _) =>
                  (checkExp e1 vtable x; checkExp e vtable x)
	      	  | _ => raise Error ("integer variable used as array",p)))
    | Janus.SubUpdate (lv,e,pos) =>
        (case lv of
           Janus.IntVar (x,p) =>
             (case lookup x vtable of
               SOME Integer =>
                 checkExp e vtable x
	      	  | _ => raise Error ("array variable used as integer",p))
        |  Janus.ArrayIndex (x,e1,p) =>
             (case lookup x vtable of
				SOME (Array _) =>
                  (checkExp e1 vtable x; checkExp e vtable x)
	      	  | _ => raise Error ("integer variable used as array",p)))
    | Janus.If (c1,s1,s2,c2,pos) =>
		(checkCond c1 vtable;
	 	checkStat s1 vtable pnames;
	 	checkStat s2 vtable pnames;
	 	checkCond c2 vtable)
    | Janus.Loop (c1,s1,s2,c2,pos) =>
		(checkCond c1 vtable;
	 	checkStat s1 vtable pnames;
	 	checkStat s2 vtable pnames;
	 	checkCond c2 vtable)
    | Janus.Skip pos => ()
    | Janus.Call (p,pos) =>
        if List.exists (fn q=>q=p) pnames
		then ()
		else raise Error ("Unknown procedure "^p,pos)
    | Janus.Uncall (p,pos) =>
        if List.exists (fn q=>q=p) pnames
		then ()
		else raise Error ("Unknown procedure "^p,pos)
    | Janus.Push (lv,pos) =>  (* ADDED *)
        (case lv of
           Janus.IntVar (x,p) => 
            (case lookup x vtable of
                SOME Integer => () (* num requires no typecheck *)
              | _ => raise Error ("Push from undefined variable",p))
         | Janus.ArrayIndex (x,e,p) =>
             (case lookup x vtable of
                SOME (Array _) => checkExp e vtable x (* avoid x in array index e *)
              | _ => raise Error ("Push from undefined variable",p))
        )
    | Janus.Pop (lv,pos) =>  (* ADDED *)
        (case lv of
           Janus.IntVar (x,p) => 
            (case lookup x vtable of
                SOME Integer => ()  (* num requires no typecheck *)
              | _ => raise Error ("Pop to undefined variable",p))
         | Janus.ArrayIndex (x,e,p) =>
             (case lookup x vtable of
                SOME (Array _) => checkExp e vtable x (* avoid x in array index e *)
              | _ => raise Error ("Pop to undefined variable",p))
        )
    | Janus.Swap pos => ()  (* ADDED - no need for type checking here *)

  fun checkDefs [] vtable = vtable
    | checkDefs (Janus.IntVarDef (x,pos)::defs) vtable =
        (case lookup x vtable of
	   NONE => checkDefs defs ((x,Integer)::vtable)
	 | SOME _ => raise Error ("Multiple declaration of "^x,pos))
    | checkDefs (Janus.ArrayVarDef (x,size,pos)::defs) vtable =
        (case lookup x vtable of
	   NONE => checkDefs defs ((x,Array size)::vtable)
	 | SOME _ => raise Error ("Multiple declaration of "^x,pos))

(* ADDED: specific function to check outdefs (to see if already defined as indef) *)

  fun checkOutDefs [] vtable vin = vtable
    | checkOutDefs (Janus.IntVarDef (x,pos)::defs) vtable vin =
      (case lookup x vtable of
		 NONE   => checkOutDefs defs ((x,Integer)::vtable) vin (* normal out variable *)
		 (* check if declared as in variable and has correct type, else error *)
	   | SOME _ =>
			(case lookupDef x vin of 
				SOME Integer =>  (* matching in/out type*)
					checkOutDefs defs ((x,Integer)::vtable) vin  
			  | SOME (Array _) =>  (* in/out type mismatch *)
					raise Error ("Non-matching type of in/out variable "^x,pos) 
			  | NONE => (* already declared as out *)
					raise Error ("Multiple declaration of "^x,pos) ) 
	  )
    | checkOutDefs (Janus.ArrayVarDef (x,size,pos)::defs) vtable vin =
	  (case lookup x vtable of
		 NONE   => checkOutDefs defs ((x,Array size)::vtable) vin (* normal out variable *)
		 (* check if declared as in variable and has correct type, else error *)
	   | SOME _ =>
			(case lookupDef x vin of 
			    SOME (Array y) => (
					if size=y then  (* matching in/out type and dimension*)
						checkOutDefs defs ((x,Array size)::vtable) vin
					else  (* dimension mismatch *)
						raise Error ("Non-matching dimension of in/out variable "^x,pos) ) 
			  | SOME Integer =>  (* in/out type mismatch *)
					raise Error ("Non-matching type of in/out variable "^x,pos) 
			  | NONE =>  (* already declared as out *)
					raise Error ("Multiple declaration of "^x,pos) ) 
	  )

  fun getProcs [] pnames = pnames
    | getProcs ((p,s,pos)::procs) pnames =
        if List.exists (fn q=>q=p) pnames
	then raise Error ("Multiply declared procedure "^p,pos)
	else getProcs procs (p::pnames)

  fun checkProcs [] vtable pnames = ()
    | checkProcs ((p,s,pos)::procs) vtable pnames =
        (checkStat s vtable pnames; checkProcs procs vtable pnames)

  (* main function for checking program *)
  fun checkProgram (ins,outs,locals,stat,procs) =
    let
      val vtable0 = checkDefs ins []
      val vtable1 = checkOutDefs outs vtable0 ins  (* modified: call checkOutDefs with ins as extra argument *)
      val vtable2 = checkDefs locals vtable1
      val pnames = getProcs procs []
      val () = checkProcs procs vtable2 pnames
    in
      checkStat stat vtable2 pnames
    end

end
